@@ -1,7 +1,9 @@
+// Copyright (c) 2021 Red Hat, Inc.
+// Copyright Contributors to the Open Cluster Management project
 import '@patternfly/react-core/dist/styles/base.css'
 import React, { Fragment } from 'react'
 import { ButtonVariant, ModalVariant } from '@patternfly/react-core'
-import { AcmAlert, AcmModal, AcmButton } from '@stolostron/ui-components'
+import { AcmAlert, AcmModal, AcmButton } from '@open-cluster-management/ui-components'
 import { useTranslation } from 'react-i18next'
 import {
     SearchResultItemsDocument,
@@ -195,11 +197,9 @@ export const DeleteResourceModal = (props: any) => {
                 variant={ModalVariant.medium}
                 isOpen={open}
                 title={t('search.modal.delete.resource.title', { resourceKind: resource?.kind })}
+                titleIconVariant={'warning'}
                 onClose={close}
                 actions={[
-                    <AcmButton key="cancel" variant={ButtonVariant.secondary} onClick={close}>
-                        {t('search.modal.delete.resource.action.cancel')}
-                    </AcmButton>,
                     <AcmButton
                         isDisabled={
                             userAccessResponse.loading ||
@@ -211,6 +211,9 @@ export const DeleteResourceModal = (props: any) => {
                     >
                         {t('search.modal.delete.resource.action.delete')}
                     </AcmButton>,
+                    <AcmButton key="cancel" variant={ButtonVariant.secondary} onClick={close}>
+                        {t('search.modal.delete.resource.action.cancel')}
+                    </AcmButton>,
                 ]}
             >
                 {userAccessResponse.error ? (
