@@ -107,7 +107,7 @@ export class MoveLinesCommand {
                     // add edit operations for moving line first to make sure it's executed after we make indentation change
                     // to s.startLineNumber
                     builder.addEditOperation(new Range(s.startLineNumber, 1, s.startLineNumber, 1), insertingText + '\n');
-                    let ret = this.matchEnterRule(model, indentConverter, tabSize, s.startLineNumber, s.startLineNumber, insertingText);
+                    let ret = this.matchEnterRuleMovingDown(model, indentConverter, tabSize, s.startLineNumber, movingLineNumber, insertingText);
                     // check if the line being moved before matches onEnter rules, if so let's adjust the indentation by onEnter rules.
                     if (ret !== null) {
                         if (ret !== 0) {
@@ -196,28 +196,7 @@ export class MoveLinesCommand {
             }
         };
     }
-    matchEnterRule(model, indentConverter, tabSize, line, oneLineAbove, oneLineAboveText) {
-        let validPrecedingLine = oneLineAbove;
-        while (validPrecedingLine >= 1) {
-            // ship empty lines as empty lines just inherit indentation
-            let lineContent;
-            if (validPrecedingLine === oneLineAbove && oneLineAboveText !== undefined) {
-                lineContent = oneLineAboveText;
-            }
-            else {
-                lineContent = model.getLineContent(validPrecedingLine);
-            }
-            let nonWhitespaceIdx = strings.lastNonWhitespaceIndex(lineContent);
-            if (nonWhitespaceIdx >= 0) {
-                break;
-            }
-            validPrecedingLine--;
-        }
-        if (validPrecedingLine < 1 || line > model.getLineCount()) {
-            return null;
-        }
-        let maxColumn = model.getLineMaxColumn(validPrecedingLine);
-        let enter = LanguageConfigurationRegistry.getEnterAction(this._autoIndent, model, new Range(validPrecedingLine, maxColumn, validPrecedingLine, maxColumn));
+    parseEnterResult(model, indentConverter, tabSize, line, enter) {
         if (enter) {
             let enterPrefix = enter.indentation;
             if (enter.indentAction === IndentAction.None) {
@@ -247,6 +226,65 @@ export class MoveLinesCommand {
         }
         return null;
     }
+    /**
+     *
+     * @param model
+     * @param indentConverter
+     * @param tabSize
+     * @param line the line moving down
+     * @param futureAboveLineNumber the line which will be at the `line` position
+     * @param futureAboveLineText
+     */
+    matchEnterRuleMovingDown(model, indentConverter, tabSize, line, futureAboveLineNumber, futureAboveLineText) {
+        if (strings.lastNonWhitespaceIndex(futureAboveLineText) >= 0) {
+            // break
+            let maxColumn = model.getLineMaxColumn(futureAboveLineNumber);
+            let enter = LanguageConfigurationRegistry.getEnterAction(this._autoIndent, model, new Range(futureAboveLineNumber, maxColumn, futureAboveLineNumber, maxColumn));
+            return this.parseEnterResult(model, indentConverter, tabSize, line, enter);
+        }
+        else {
+            // go upwards, starting from `line - 1`
+            let validPrecedingLine = line - 1;
+            while (validPrecedingLine >= 1) {
+                let lineContent = model.getLineContent(validPrecedingLine);
+                let nonWhitespaceIdx = strings.lastNonWhitespaceIndex(lineContent);
+                if (nonWhitespaceIdx >= 0) {
+                    break;
+                }
+                validPrecedingLine--;
+            }
+            if (validPrecedingLine < 1 || line > model.getLineCount()) {
+                return null;
+            }
+            let maxColumn = model.getLineMaxColumn(validPrecedingLine);
+            let enter = LanguageConfigurationRegistry.getEnterAction(this._autoIndent, model, new Range(validPrecedingLine, maxColumn, validPrecedingLine, maxColumn));
+            return this.parseEnterResult(model, indentConverter, tabSize, line, enter);
+        }
+    }
+    matchEnterRule(model, indentConverter, tabSize, line, oneLineAbove, previousLineText) {
+        let validPrecedingLine = oneLineAbove;
+        while (validPrecedingLine >= 1) {
+            // ship empty lines as empty lines just inherit indentation
+            let lineContent;
+            if (validPrecedingLine === oneLineAbove && previousLineText !== undefined) {
+                lineContent = previousLineText;
+            }
+            else {
+                lineContent = model.getLineContent(validPrecedingLine);
+            }
+            let nonWhitespaceIdx = strings.lastNonWhitespaceIndex(lineContent);
+            if (nonWhitespaceIdx >= 0) {
+                break;
+            }
+            validPrecedingLine--;
+        }
+        if (validPrecedingLine < 1 || line > model.getLineCount()) {
+            return null;
+        }
+        let maxColumn = model.getLineMaxColumn(validPrecedingLine);
+        let enter = LanguageConfigurationRegistry.getEnterAction(this._autoIndent, model, new Range(validPrecedingLine, maxColumn, validPrecedingLine, maxColumn));
+        return this.parseEnterResult(model, indentConverter, tabSize, line, enter);
+    }
     trimLeft(str) {
         return str.replace(/^\s+/, '');
     }
