@@ -1,6 +1,5 @@
 'use strict'
 
-const { URL } = require('url')
 const net = require('net')
 const tls = require('tls')
 const HTTPParser = require('../node/http-parser')
@@ -10,22 +9,20 @@ const util = require('./util')
 const Request = require('./request')
 const {
   ContentLengthMismatchError,
-  SocketTimeoutError,
   TrailerMismatchError,
   InvalidArgumentError,
   RequestAbortedError,
+  HeadersTimeoutError,
   ClientDestroyedError,
   ClientClosedError,
-  HeadersTimeoutError,
   SocketError,
-  InformationalError
+  InformationalError,
+  BodyTimeoutError
 } = require('./errors')
 const {
   kUrl,
   kReset,
-  kPause,
   kHost,
-  kResume,
   kClient,
   kParser,
   kConnect,
@@ -34,8 +31,8 @@ const {
   kQueue,
   kNeedDrain,
   kTLSServerName,
-  kIdleTimeout,
-  kSocketTimeout,
+  kKeepAliveDefaultTimeout,
+  kHostHeader,
   kTLSOpts,
   kClosed,
   kDestroyed,
@@ -44,21 +41,19 @@ const {
   kError,
   kOnDestroyed,
   kPipelining,
-  kRetryDelay,
-  kRetryTimeout,
   kSocket,
   kSocketPath,
-  kKeepAliveTimeout,
+  kKeepAliveTimeoutValue,
   kMaxHeadersSize,
-  kHeadersTimeout,
   kKeepAliveMaxTimeout,
   kKeepAliveTimeoutThreshold,
-  kKeepAlive,
-  kTLSSession
+  kTLSSession,
+  kIdleTimeout,
+  kIdleTimeoutValue,
+  kHeadersTimeout,
+  kBodyTimeout
 } = require('./symbols')
 
-const kHostHeader = Symbol('host header')
-
 const nodeVersions = process.version.split('.')
 const nodeMajorVersion = parseInt(nodeVersions[0].slice(1))
 const nodeMinorVersion = parseInt(nodeVersions[1])
@@ -78,10 +73,13 @@ class Client extends EventEmitter {
     maxHeaderSize,
     headersTimeout,
     socketTimeout,
+    requestTimeout,
+    bodyTimeout,
     idleTimeout,
-    maxKeepAliveTimeout,
     keepAlive,
-    keepAliveMaxTimeout = maxKeepAliveTimeout,
+    keepAliveTimeout,
+    maxKeepAliveTimeout,
+    keepAliveMaxTimeout,
     keepAliveTimeoutThreshold,
     socketPath,
     pipelining,
@@ -89,93 +87,75 @@ class Client extends EventEmitter {
   } = {}) {
     super()
 
-    if (typeof url === 'string') {
-      url = new URL(url)
+    if (keepAlive !== undefined) {
+      throw new InvalidArgumentError('unsupported keepAlive, use pipelining=0 instead')
     }
 
-    if (!url || typeof url !== 'object') {
-      throw new InvalidArgumentError('invalid url')
+    if (socketTimeout !== undefined) {
+      throw new InvalidArgumentError('unsupported socketTimeout, use headersTimeout & bodyTimeout instead')
     }
 
-    if (url.port != null && url.port !== '' && !Number.isFinite(parseInt(url.port))) {
-      throw new InvalidArgumentError('invalid port')
+    if (requestTimeout !== undefined) {
+      throw new InvalidArgumentError('unsupported requestTimeout, use headersTimeout & bodyTimeout instead')
     }
 
-    if (socketPath != null && typeof socketPath !== 'string') {
-      throw new InvalidArgumentError('invalid socketPath')
+    if (idleTimeout !== undefined) {
+      throw new InvalidArgumentError('unsupported idleTimeout, use keepAliveTimeout instead')
     }
 
-    if (url.hostname != null && typeof url.hostname !== 'string') {
-      throw new InvalidArgumentError('invalid hostname')
-    }
-
-    if (!/https?/.test(url.protocol)) {
-      throw new InvalidArgumentError('invalid protocol')
-    }
-
-    if (/\/.+/.test(url.pathname) || url.search || url.hash) {
-      throw new InvalidArgumentError('invalid url')
+    if (maxKeepAliveTimeout !== undefined) {
+      throw new InvalidArgumentError('unsupported maxKeepAliveTimeout, use keepAliveMaxTimeout instead')
     }
 
     if (maxHeaderSize != null && !Number.isFinite(maxHeaderSize)) {
       throw new InvalidArgumentError('invalid maxHeaderSize')
     }
 
-    if (socketTimeout != null && !Number.isFinite(socketTimeout)) {
-      throw new InvalidArgumentError('invalid socketTimeout')
+    if (socketPath != null && typeof socketPath !== 'string') {
+      throw new InvalidArgumentError('invalid socketPath')
     }
 
-    if (idleTimeout != null && (!Number.isFinite(idleTimeout) || idleTimeout <= 0)) {
-      throw new InvalidArgumentError('invalid idleTimeout')
+    if (keepAliveTimeout != null && (!Number.isFinite(keepAliveTimeout) || keepAliveTimeout <= 0)) {
+      throw new InvalidArgumentError('invalid keepAliveTimeout')
     }
 
     if (keepAliveMaxTimeout != null && (!Number.isFinite(keepAliveMaxTimeout) || keepAliveMaxTimeout <= 0)) {
       throw new InvalidArgumentError('invalid keepAliveMaxTimeout')
     }
 
-    if (keepAlive != null && typeof keepAlive !== 'boolean') {
-      throw new InvalidArgumentError('invalid keepAlive')
-    }
-
     if (keepAliveTimeoutThreshold != null && !Number.isFinite(keepAliveTimeoutThreshold)) {
       throw new InvalidArgumentError('invalid keepAliveTimeoutThreshold')
     }
 
-    if (headersTimeout != null && !Number.isFinite(headersTimeout)) {
-      throw new InvalidArgumentError('invalid headersTimeout')
+    if (headersTimeout != null && (!Number.isInteger(headersTimeout) || headersTimeout < 0)) {
+      throw new InvalidArgumentError('headersTimeout must be a positive integer or zero')
+    }
+
+    if (bodyTimeout != null && (!Number.isInteger(bodyTimeout) || bodyTimeout < 0)) {
+      throw new InvalidArgumentError('bodyTimeout must be a positive integer or zero')
     }
 
     this[kSocket] = null
-    this[kReset] = false
-    this[kPipelining] = pipelining || 1
+    this[kPipelining] = pipelining != null ? pipelining : 1
     this[kMaxHeadersSize] = maxHeaderSize || 16384
-    this[kHeadersTimeout] = headersTimeout == null ? 30e3 : headersTimeout
-    this[kUrl] = url
+    this[kUrl] = util.parseOrigin(url)
     this[kSocketPath] = socketPath
-    this[kSocketTimeout] = socketTimeout == null ? 30e3 : socketTimeout
-    this[kIdleTimeout] = idleTimeout == null ? 4e3 : idleTimeout
+    this[kKeepAliveDefaultTimeout] = keepAliveTimeout == null ? 4e3 : keepAliveTimeout
     this[kKeepAliveMaxTimeout] = keepAliveMaxTimeout == null ? 600e3 : keepAliveMaxTimeout
     this[kKeepAliveTimeoutThreshold] = keepAliveTimeoutThreshold == null ? 1e3 : keepAliveTimeoutThreshold
-    this[kKeepAliveTimeout] = this[kIdleTimeout]
-    this[kKeepAlive] = keepAlive == null ? true : keepAlive
+    this[kKeepAliveTimeoutValue] = this[kKeepAliveDefaultTimeout]
     this[kClosed] = false
     this[kDestroyed] = false
     this[kTLSOpts] = tls
     this[kTLSServerName] = getServerName(this)
     this[kHost] = null
-    this[kRetryDelay] = 0
-    this[kRetryTimeout] = null
     this[kOnDestroyed] = []
-    this[kWriting] = false
     this[kResuming] = 0 // 0, idle, 1, scheduled, 2 resuming
     this[kNeedDrain] = 0 // 0, idle, 1, scheduled, 2 resuming
-    this[kResume] = () => {
-      if (this[kResuming] === 0) {
-        resume(this, true)
-      }
-    }
     this[kTLSSession] = null
     this[kHostHeader] = `host: ${this[kUrl].hostname}${this[kUrl].port ? `:${this[kUrl].port}` : ''}\r\n`
+    this[kBodyTimeout] = bodyTimeout != null ? bodyTimeout : 30e3
+    this[kHeadersTimeout] = headersTimeout != null ? headersTimeout : 30e3
 
     // kQueue is built up of 3 sections separated by
     // the kRunningIdx and kPendingIdx indices.
@@ -191,6 +171,10 @@ class Client extends EventEmitter {
     this[kPendingIdx] = 0
   }
 
+  get url () {
+    return this[kUrl]
+  }
+
   get pipelining () {
     return this[kPipelining]
   }
@@ -225,7 +209,12 @@ class Client extends EventEmitter {
   }
 
   get busy () {
-    return this[kReset] || this[kWriting] || this.pending > 0
+    const socket = this[kSocket]
+    return (
+      (socket && (socket[kReset] || socket[kWriting])) ||
+      (this.size >= (this[kPipelining] || 1)) ||
+      this.pending > 0
+    )
   }
 
   get destroyed () {
@@ -264,6 +253,10 @@ class Client extends EventEmitter {
         resume(this, true)
       }
     } catch (err) {
+      if (typeof handler.onError !== 'function') {
+        throw new InvalidArgumentError('invalid onError method')
+      }
+
       handler.onError(err)
     }
   }
@@ -330,8 +323,6 @@ class Client extends EventEmitter {
       request.onError(err)
     }
 
-    clearTimeout(this[kRetryTimeout])
-    this[kRetryTimeout] = null
     this[kClosed] = true
     this[kDestroyed] = true
     this[kOnDestroyed].push(callback)
@@ -362,7 +353,7 @@ class Parser extends HTTPParser {
       this.initialize(
         HTTPParser.RESPONSE,
         {},
-        client[kHeadersTimeout]
+        0
       )
     } else if (nodeMajorVersion === 12 && nodeMinorVersion >= 19) {
       super()
@@ -370,42 +361,85 @@ class Parser extends HTTPParser {
         HTTPParser.RESPONSE,
         {},
         client[kMaxHeadersSize],
-        client[kHeadersTimeout]
+        0
       )
-    } else if (nodeMajorVersion > 12) {
+    } else if (nodeMajorVersion > 12 && nodeMajorVersion < 16) {
       super()
       this.initialize(
         HTTPParser.RESPONSE,
         {},
         client[kMaxHeadersSize],
         insecureHTTPParser,
-        client[kHeadersTimeout]
+        0
+      )
+    } else if (nodeMajorVersion >= 16) {
+      super()
+      this.initialize(
+        HTTPParser.RESPONSE,
+        {},
+        client[kMaxHeadersSize],
+        0
       )
     } else {
-      super(HTTPParser.RESPONSE, false)
+      super(HTTPParser.RESPONSE)
     }
 
     this.client = client
     this.socket = socket
-
+    this.timeout = null
     this.statusCode = null
     this.upgrade = false
     this.headers = null
     this.shouldKeepAlive = false
-    this.read = 0
     this.request = null
-  }
+    this.paused = false
 
-  [HTTPParser.kOnTimeout] () {
-    /* istanbul ignore next: https://github.com/nodejs/node/pull/34578 */
-    if (this.statusCode) {
-      this.socket._unrefTimer()
-    } else {
-      util.destroy(this.socket, new HeadersTimeoutError())
+    // Parser can't be paused from within a callback.
+    // Use a buffer in JS land in order to stop further progress while paused.
+    this.resuming = false
+    this.queue = []
+
+    this._resume = () => {
+      if (!this.paused || this.resuming) {
+        return
+      }
+
+      this.paused = false
+
+      this.resuming = true
+      while (this.queue.length) {
+        const [fn, ...args] = this.queue.shift()
+
+        Reflect.apply(fn, this, args)
+
+        if (this.paused) {
+          this.resuming = false
+          return
+        }
+      }
+      this.resuming = false
+
+      socketResume(this.socket)
+    }
+
+    this._pause = () => {
+      if (this.paused) {
+        return
+      }
+
+      this.paused = true
+
+      socketPause(this.socket)
     }
   }
 
   [HTTPParser.kOnHeaders] (rawHeaders) {
+    /* istanbul ignore next: difficult to make a test case for */
+    if (this.paused) {
+      this.queue.push([this[HTTPParser.kOnHeaders], rawHeaders])
+      return
+    }
+
     if (this.headers) {
       Array.prototype.push.apply(this.headers, rawHeaders)
     } else {
@@ -414,6 +448,11 @@ class Parser extends HTTPParser {
   }
 
   [HTTPParser.kOnExecute] (ret) {
+    if (this.paused) {
+      this.queue.push([this[HTTPParser.kOnExecute], ret])
+      return
+    }
+
     const { upgrade, socket } = this
 
     if (!Number.isFinite(ret)) {
@@ -422,12 +461,6 @@ class Parser extends HTTPParser {
       return
     }
 
-    // When the underlying `net.Socket` instance is consumed - no
-    // `data` events are emitted, and thus `socket.setTimeout` fires the
-    // callback even if the data is constantly flowing into the socket.
-    // See, https://github.com/nodejs/node/commit/ec2822adaad76b126b5cccdeaa1addf2376c9aa6
-    socket._unrefTimer()
-
     // This logic cannot live in kOnHeadersComplete since we
     // have no way of slicing the parsing buffer without knowing
     // the offset which is only provided in kOnExecute.
@@ -437,7 +470,8 @@ class Parser extends HTTPParser {
       assert(!socket.destroyed)
       assert(socket === client[kSocket])
       assert(!socket.isPaused())
-      assert(request.upgrade || request.method === 'CONNECT')
+      assert(socket._handle && socket._handle.reading)
+      assert(request.upgrade)
 
       this.headers = null
       this.statusCode = null
@@ -458,27 +492,31 @@ class Parser extends HTTPParser {
 
       try {
         request.onUpgrade(statusCode, headers, socket)
-      } catch (err) {
-        util.destroy(socket, err)
-        request.onError(err)
-      }
 
-      if (!socket.destroyed && !request.aborted) {
-        detachSocket(socket)
+        if (!socket.destroyed && !request.aborted) {
+          detachSocket(socket)
+          client[kSocket] = null
 
-        client[kSocket] = null
-        client[kQueue][client[kRunningIdx]++] = null
-        client.emit('disconnect', new InformationalError('upgrade'))
-
-        setImmediate(() => this.close())
+          client[kQueue][client[kRunningIdx]++] = null
+          client.emit('disconnect', new InformationalError('upgrade'))
+        }
 
         resume(client)
+      } catch (err) {
+        util.destroy(socket, err)
       }
     }
   }
 
   [HTTPParser.kOnHeadersComplete] (versionMajor, versionMinor, rawHeaders, method,
     url, statusCode, statusMessage, upgrade, shouldKeepAlive) {
+    /* istanbul ignore next: difficult to make a test case for */
+    if (this.paused) {
+      this.queue.push([this[HTTPParser.kOnHeadersComplete], versionMajor, versionMinor, rawHeaders, method,
+        url, statusCode, statusMessage, upgrade, shouldKeepAlive])
+      return
+    }
+
     const { client, socket } = this
 
     const request = client[kQueue][client[kRunningIdx]]
@@ -488,6 +526,11 @@ class Parser extends HTTPParser {
       return
     }
 
+    clearTimeout(this.timeout)
+    this.timeout = client[kBodyTimeout]
+      ? setTimeout(onBodyTimeout, client[kBodyTimeout], this)
+      : null
+
     // TODO: Check for content-length mismatch from server?
 
     assert(!this.upgrade)
@@ -500,7 +543,7 @@ class Parser extends HTTPParser {
       return 1
     }
 
-    if (upgrade !== Boolean(request.upgrade)) {
+    if (request.upgrade !== true && upgrade !== Boolean(request.upgrade)) {
       util.destroy(socket, new SocketError('bad upgrade'))
       return 1
     }
@@ -515,7 +558,7 @@ class Parser extends HTTPParser {
     this.shouldKeepAlive = shouldKeepAlive
     this.request = request
 
-    if (upgrade || request.method === 'CONNECT') {
+    if (request.upgrade) {
       this.unconsume()
       this.upgrade = true
       return 2
@@ -524,9 +567,12 @@ class Parser extends HTTPParser {
     let keepAlive
     let trailers
 
-    for (let n = 0; n < this.headers.length; n += 2) {
-      const key = this.headers[n + 0]
-      const val = this.headers[n + 1]
+    const { headers } = this
+    this.headers = null
+
+    for (let n = 0; n < headers.length; n += 2) {
+      const key = headers[n + 0]
+      const val = headers[n + 1]
 
       if (!keepAlive && key.length === 10 && key.toLowerCase() === 'keep-alive') {
         keepAlive = val
@@ -535,11 +581,9 @@ class Parser extends HTTPParser {
       }
     }
 
-    const { headers } = this
-    this.headers = null
     this.trailers = trailers ? trailers.toLowerCase().split(/,\s*/) : null
 
-    if (shouldKeepAlive && client[kKeepAlive]) {
+    if (shouldKeepAlive && client[kPipelining]) {
       const keepAliveTimeout = keepAlive ? util.parseKeepAliveTimeout(keepAlive) : null
 
       if (keepAliveTimeout != null) {
@@ -547,22 +591,22 @@ class Parser extends HTTPParser {
           keepAliveTimeout - client[kKeepAliveTimeoutThreshold],
           client[kKeepAliveMaxTimeout]
         )
-        if (timeout < 1e3) {
-          client[kReset] = true
+        if (timeout <= 0) {
+          socket[kReset] = true
         } else {
-          client[kKeepAliveTimeout] = timeout
+          client[kKeepAliveTimeoutValue] = timeout
         }
       } else {
-        client[kKeepAliveTimeout] = client[kIdleTimeout]
+        client[kKeepAliveTimeoutValue] = client[kKeepAliveDefaultTimeout]
       }
     } else {
       // Stop more requests from being dispatched.
-      client[kReset] = true
+      socket[kReset] = true
     }
 
     try {
-      if (request.onHeaders(statusCode, headers, socket[kResume]) === false) {
-        socket[kPause]()
+      if (request.onHeaders(statusCode, headers, this._resume) === false) {
+        this._pause()
       }
     } catch (err) {
       util.destroy(socket, err)
@@ -573,17 +617,26 @@ class Parser extends HTTPParser {
   }
 
   [HTTPParser.kOnBody] (chunk, offset, length) {
-    const { socket, statusCode, request } = this
+    if (this.paused) {
+      this.queue.push([this[HTTPParser.kOnBody], chunk, offset, length])
+      return
+    }
+
+    const { socket, statusCode, request, timeout } = this
 
     if (socket.destroyed) {
       return
     }
 
+    if (timeout && timeout.refresh) {
+      timeout.refresh()
+    }
+
     assert(statusCode >= 200)
 
     try {
-      if (request.onBody(chunk, offset, length) === false) {
-        socket[kPause]()
+      if (request.onData(chunk.slice(offset, offset + length)) === false) {
+        this._pause()
       }
     } catch (err) {
       util.destroy(socket, err)
@@ -591,6 +644,12 @@ class Parser extends HTTPParser {
   }
 
   [HTTPParser.kOnMessageComplete] () {
+    /* istanbul ignore next: difficult to make a test case for */
+    if (this.paused) {
+      this.queue.push([this[HTTPParser.kOnMessageComplete]])
+      return
+    }
+
     const { client, socket, statusCode, headers, upgrade, request, trailers } = this
 
     if (socket.destroyed) {
@@ -598,6 +657,7 @@ class Parser extends HTTPParser {
     }
 
     assert(statusCode >= 100)
+    assert(this.resuming || (socket._handle && socket._handle.reading))
 
     if (upgrade) {
       // TODO: When, how and why does this happen?
@@ -610,8 +670,12 @@ class Parser extends HTTPParser {
     this.request = null
     this.trailers = null
 
+    clearTimeout(this.timeout)
+    this.timeout = client[kHeadersTimeout]
+      ? setTimeout(onHeadersTimeout, client[kHeadersTimeout], this)
+      : null
+
     if (statusCode < 200) {
-      assert(!socket.isPaused())
       return
     }
 
@@ -636,7 +700,11 @@ class Parser extends HTTPParser {
         }
       }
 
-      request.onComplete(headers)
+      try {
+        request.onComplete(headers)
+      } catch (err) {
+        request.onError(err)
+      }
     } catch (err) {
       util.destroy(socket, err)
       return
@@ -644,46 +712,47 @@ class Parser extends HTTPParser {
 
     client[kQueue][client[kRunningIdx]++] = null
 
-    if (client[kWriting]) {
+    if (socket[kWriting]) {
       // Response completed before request.
       util.destroy(socket, new InformationalError('reset'))
     } else if (!this.shouldKeepAlive) {
       util.destroy(socket, new InformationalError('reset'))
-    } else if (client[kReset] && !client.running) {
+    } else if (socket[kReset] && !client.running) {
       // Destroy socket once all requests have completed.
       // The request at the tail of the pipeline is the one
       // that requested reset and no further requests should
       // have been queued since then.
       util.destroy(socket, new InformationalError('reset'))
     } else {
-      socket[kResume]()
       resume(client)
     }
   }
 
   destroy () {
+    clearTimeout(this.timeout)
+    this.timeout = null
     this.unconsume()
     setImmediate((self) => self.close(), this)
   }
 }
 
+function onBodyTimeout (self) {
+  if (!self.paused) {
+    util.destroy(self.socket, new BodyTimeoutError())
+  }
+}
+
+function onHeadersTimeout (self) {
+  util.destroy(self.socket, new HeadersTimeoutError())
+}
+
 function onSocketConnect () {
   const { [kClient]: client } = this
 
-  assert(!this.destroyed)
-  assert(!client[kWriting])
-  assert(!client[kRetryTimeout])
-
-  client[kReset] = false
-  client[kRetryDelay] = 0
   client.emit('connect')
   resume(client)
 }
 
-function onSocketTimeout () {
-  util.destroy(this, new SocketTimeoutError())
-}
-
 function onSocketError (err) {
   const { [kClient]: client } = this
 
@@ -712,15 +781,31 @@ function onSocketEnd () {
   util.destroy(this, new SocketError('other side closed'))
 }
 
+function detachSocket (socket) {
+  clearTimeout(socket[kIdleTimeout])
+  socket[kIdleTimeout] = null
+  socket[kIdleTimeoutValue] = null
+
+  socket[kParser].destroy()
+  socket[kParser] = null
+
+  socket[kClient] = null
+  socket[kError] = null
+  socket
+    .removeListener('session', onSocketSession)
+    .removeListener('error', onSocketError)
+    .removeListener('end', onSocketEnd)
+    .removeListener('close', onSocketClose)
+}
+
 function onSocketClose () {
-  const { [kClient]: client, [kParser]: parser } = this
+  const { [kClient]: client } = this
 
   const err = this[kError] || new SocketError('closed')
 
+  detachSocket(this)
   client[kSocket] = null
 
-  parser.destroy()
-
   if (err.code !== 'UND_ERR_INFO') {
     // Evict session on errors.
     client[kTLSSession] = null
@@ -756,24 +841,8 @@ function onSocketSession (session) {
   client[kTLSSession] = session
 }
 
-function detachSocket (socket) {
-  socket[kParser].destroy()
-  socket[kParser] = null
-  socket[kPause] = null
-  socket[kResume] = null
-  socket[kClient] = null
-  socket[kError] = null
-  socket
-    .removeListener('timeout', onSocketTimeout)
-    .removeListener('session', onSocketSession)
-    .removeListener('error', onSocketError)
-    .removeListener('end', onSocketEnd)
-    .removeListener('close', onSocketClose)
-}
-
 function connect (client) {
   assert(!client[kSocket])
-  assert(!client[kRetryTimeout])
 
   const { protocol, port, hostname } = client[kUrl]
 
@@ -810,39 +879,37 @@ function connect (client) {
     parser.consume(socket._handle._externalStream)
   }
 
-  socket[kPause] = socketPause.bind(socket)
-  socket[kResume] = socketResume.bind(socket)
+  socket[kIdleTimeout] = null
+  socket[kIdleTimeoutValue] = null
+  socket[kWriting] = false
+  socket[kReset] = false
   socket[kError] = null
   socket[kParser] = parser
   socket[kClient] = client
   socket
     .setNoDelay(true)
-    .setTimeout(client[kIdleTimeout])
     .on(protocol === 'https:' ? 'secureConnect' : 'connect', onSocketConnect)
-    .on('timeout', onSocketTimeout)
     .on('error', onSocketError)
     .on('end', onSocketEnd)
     .on('close', onSocketClose)
 }
 
-function socketPause () {
-  // TODO: Pause parser.
-  if (this._handle && this._handle.reading) {
-    this._handle.reading = false
-    const err = this._handle.readStop()
+function socketPause (socket) {
+  if (socket._handle && socket._handle.reading) {
+    socket._handle.reading = false
+    const err = socket._handle.readStop()
     if (err) {
-      this.destroy(util.errnoException(err, 'read'))
+      socket.destroy(util.errnoException(err, 'read'))
     }
   }
 }
 
-function socketResume () {
-  // TODO: Resume parser.
-  if (this._handle && !this._handle.reading) {
-    this._handle.reading = true
-    const err = this._handle.readStart()
+function socketResume (socket) {
+  if (socket._handle && !socket._handle.reading) {
+    socket._handle.reading = true
+    const err = socket._handle.readStart()
     if (err) {
-      this.destroy(util.errnoException(err, 'read'))
+      socket.destroy(util.errnoException(err, 'read'))
     }
   }
 }
@@ -881,12 +948,17 @@ function _resume (client, sync) {
     }
 
     if (client[kSocket]) {
-      const timeout = client.running
-        ? client[kSocketTimeout]
-        : client[kKeepAliveTimeout]
+      const socket = client[kSocket]
+      const timeout = client.running ? 0 : client[kKeepAliveTimeoutValue]
 
-      if (client[kSocket].timeout !== timeout) {
-        client[kSocket].setTimeout(timeout)
+      if (socket[kIdleTimeoutValue] !== timeout) {
+        clearTimeout(socket[kIdleTimeout])
+        if (timeout) {
+          socket[kIdleTimeout] = setTimeout((socket) => {
+            util.destroy(socket, new InformationalError('socket idle timeout'))
+          }, timeout, socket)
+        }
+        socket[kIdleTimeoutValue] = timeout
       }
     }
 
@@ -898,25 +970,27 @@ function _resume (client, sync) {
       }
     }
 
-    if (!client.pending) {
-      if (client[kNeedDrain] === 2 && !client.busy) {
-        if (sync) {
-          client[kNeedDrain] = 1
-          process.nextTick(emitDrain, client)
-        } else {
-          emitDrain(client)
-        }
-        continue
-      }
-      return
-    } else {
+    if (client.busy) {
       client[kNeedDrain] = 2
+    } else if (client[kNeedDrain] === 2) {
+      if (sync) {
+        client[kNeedDrain] = 1
+        process.nextTick(emitDrain, client)
+      } else {
+        emitDrain(client)
+      }
+      continue
     }
 
-    if (client.running >= client[kPipelining]) {
+    if (!client.pending) {
       return
     }
 
+    if (client.running >= (client[kPipelining] || 1)) {
+      return
+    }
+
+    const socket = client[kSocket]
     const request = client[kQueue][client[kPendingIdx]]
 
     if (client[kUrl].protocol === 'https:' && client[kHost] !== request.host) {
@@ -932,24 +1006,15 @@ function _resume (client, sync) {
         client[kTLSServerName] = servername
         client[kTLSSession] = null
 
-        if (client[kSocket]) {
-          util.destroy(client[kSocket], new InformationalError('servername changed'))
+        if (socket) {
+          util.destroy(socket, new InformationalError('servername changed'))
           return
         }
       }
     }
 
-    if (!client[kSocket] && !client[kRetryTimeout]) {
-      if (client[kRetryDelay]) {
-        client[kRetryTimeout] = setTimeout(() => {
-          client[kRetryTimeout] = null
-          connect(client)
-        }, client[kRetryDelay])
-        client[kRetryDelay] = Math.min(client[kRetryDelay] * 2, client[kSocketTimeout])
-      } else {
-        connect(client)
-        client[kRetryDelay] = 1e3
-      }
+    if (!socket) {
+      connect(client)
       return
     }
 
@@ -957,12 +1022,7 @@ function _resume (client, sync) {
       return
     }
 
-    if (client[kWriting] || client[kReset]) {
-      return
-    }
-
-    if (client.running && !client[kKeepAlive]) {
-      // Don't schedule more if we know connection will reset.
+    if (socket[kWriting] || socket[kReset]) {
       return
     }
 
@@ -973,7 +1033,7 @@ function _resume (client, sync) {
       return
     }
 
-    if (client.running && (request.upgrade || request.method === 'CONNECT')) {
+    if (client.running && request.upgrade) {
       // Don't dispatch an upgrade until all preceeding requests have completed.
       // A misbehaving server might upgrade the connection before all pipelined
       // request has completed.
@@ -1008,7 +1068,12 @@ function _resume (client, sync) {
       return
     }
 
-    if (write(client, request)) {
+    if (!request.aborted && write(client, request)) {
+      const parser = client[kSocket][kParser]
+      if (!parser.timeout && client[kHeadersTimeout]) {
+        parser.timeout = setTimeout(onHeadersTimeout, client[kHeadersTimeout], parser)
+      }
+
       client[kPendingIdx]++
     } else {
       client[kQueue].splice(client[kPendingIdx], 1)
@@ -1059,31 +1124,46 @@ function write (client, request) {
     return false
   }
 
+  if (request.aborted) {
+    return false
+  }
+
   try {
-    request.onConnect(client[kResume])
+    request.onConnect((err) => {
+      if (request.aborted) {
+        return
+      }
+
+      request.onError(err || new RequestAbortedError())
+
+      if (client[kResuming] === 0) {
+        resume(client, true)
+      }
+    })
   } catch (err) {
     request.onError(err)
-    return false
   }
 
   if (request.aborted) {
     return false
   }
 
+  const socket = client[kSocket]
+
   if (method === 'HEAD') {
     // https://github.com/mcollina/undici/issues/258
 
     // Close after a HEAD request to interop with misbehaving servers
     // that may send a body in the response.
 
-    client[kReset] = true
+    socket[kReset] = true
   }
 
-  if (method === 'CONNECT' || upgrade) {
+  if (upgrade) {
     // On CONNECT or upgrade, block pipeline from dispatching further
     // requests on this connection.
 
-    client[kReset] = true
+    socket[kReset] = true
   }
 
   // TODO: Expect: 100-continue
@@ -1094,9 +1174,9 @@ function write (client, request) {
 
   let header
 
-  if (upgrade) {
+  if (typeof upgrade === 'string') {
     header = `${method} ${path} HTTP/1.1\r\nconnection: upgrade\r\nupgrade: ${upgrade}\r\n`
-  } else if (client[kKeepAlive]) {
+  } else if (client[kPipelining]) {
     header = `${method} ${path} HTTP/1.1\r\nconnection: keep-alive\r\n`
   } else {
     header = `${method} ${path} HTTP/1.1\r\nconnection: close\r\n`
@@ -1110,8 +1190,6 @@ function write (client, request) {
     header += headers
   }
 
-  const socket = client[kSocket]
-
   if (!body) {
     if (contentLength === 0) {
       socket.write(`${header}content-length: ${contentLength}\r\n\r\n\r\n`, 'ascii')
@@ -1129,10 +1207,10 @@ function write (client, request) {
     socket.uncork()
 
     if (!expectsPayload) {
-      client[kReset] = true
+      socket[kReset] = true
     }
   } else {
-    client[kWriting] = true
+    socket[kWriting] = true
 
     assert(util.isStream(body))
     assert(contentLength !== 0 || !client.running, 'stream body cannot be pipelined')
@@ -1158,7 +1236,7 @@ function write (client, request) {
 
         if (bytesWritten === 0) {
           if (!expectsPayload) {
-            client[kReset] = true
+            socket[kReset] = true
           }
 
           if (contentLength === null) {
@@ -1198,8 +1276,8 @@ function write (client, request) {
 
       finished = true
 
-      assert(client[kWriting] && client.running <= 1)
-      client[kWriting] = false
+      assert(socket.destroyed || (socket[kWriting] && client.running <= 1))
+      socket[kWriting] = false
 
       if (!err && contentLength !== null && bytesWritten !== contentLength) {
         err = new ContentLengthMismatchError()
@@ -1214,11 +1292,14 @@ function write (client, request) {
         .removeListener('error', onFinished)
         .removeListener('close', onAbort)
 
-      request.body = null
       util.destroy(body, err)
 
       if (err) {
+        assert(client.running <= 1, 'pipeline should only contain this request')
         util.destroy(socket, err)
+      }
+
+      if (socket.destroyed) {
         return
       }
 
